from typing import Iterable, Optional, Any
from pathlib import Path

from curricula.library.configurable import Configurable, none
from curricula.library.process import Runtime
from ..grader.task import Error
from . import Executor, Connector


class ProcessExecutor(Configurable, Executor):
    """Meant for reading and comparing stdout."""

    executable_name: str
    args: Iterable[str]
    stdin: Optional[bytes]
    timeout: Optional[float]
    cwd: Optional[Path]
    stream: str

    def __init__(
            self,
            *,
            executable_name: str = none,
            args: Iterable[str] = none,
            stdin: bytes = none,
            timeout: float = none,
            cwd: Path = none,
            stream: str = none,
            **kwargs):
        """Save container information, call super."""

        super().__init__(**kwargs)

        self.executable_name = executable_name
        self.args = args
        self.stdin = stdin
        self.timeout = timeout
        self.cwd = cwd
        self.stream = stream

    def execute(self) -> Runtime:
        """Check that it ran correctly, then run the test."""

        executable = self.resources[self.executable_name]
        runtime = executable.execute(
            *self.resolve("args", default=()),
            stdin=self.resolve("stdin", default=None),
            timeout=self.resolve("timeout", default=None),
            cwd=self.resolve("cwd", default=None))
        self.details["runtime"] = runtime.dump()
        return runtime


class ProcessInputFileMixin(Configurable):
    """Enables the use of an input file rather than stdin string."""

    input_file_path: Path

    def __init__(self, *, input_file_path: Path = none, **kwargs):
        """Set stdin to the contents of the file."""

        super().__init__(**kwargs)
        self.input_file_path = input_file_path


class ProcessExitCodeConnector(Connector):
    """Output is exit code instead of stdout."""

    def connect(self, runtime: Runtime) -> int:
        """Return exit code."""

        self.details.update(runtime=runtime.dump())
        return runtime.code


class ProcessStreamConnector(Configurable, Connector):
    """Output is stdout or stderr."""

    STDOUT = "stdout"
    STDERR = "stderr"

    stream: str

    def connect(self, runtime: Runtime) -> bytes:
        """Return exit code."""

        self.details.update(runtime=runtime.dump())
        stream = self.resolve("stream", default=self.STDOUT)
        if stream == self.STDOUT:
            return runtime.stdout
        elif stream == self.STDERR:
            return runtime.stderr
        raise ValueError(f"invalid stream type specified in connector: {stream}")


class OutputFileConnector(Configurable, Connector):
    """Enables the use of an input file rather than stdin string."""

    output_file_path: Path

    def __init__(self, *, output_file_path: Path = none, **kwargs):
        """Set stdin."""

        super().__init__(**kwargs)
        self.output_file_path = output_file_path

    def connect(self, result: Any) -> bytes:
        """Call super because it might do something."""

        output_file_path = self.resolve("output_file_path")
        if not output_file_path.is_file():
            raise self.result_type(
                complete=True,
                passing=False,
                error=Error(description="no output file", suggestion=f"expected path {output_file_path}"))
        return output_file_path.read_bytes()


class VerifyRuntimeConnector(Connector):
    """Determine whether a runtime succeeded."""

    def connect(self, runtime: Runtime) -> Runtime:
        """See if the runtime raised exceptions or returned status code."""

        if runtime.raised_exception:
            error = Error(description=runtime.exception.description)
            raise self.result_type(complete=True, passing=False, error=error, details=dict(runtime=runtime.dump()))
        if runtime.timed_out:
            error = Error(
                description="timed out",
                suggestion=f"exceeded maximum elapsed time of {runtime.timeout} seconds")
            raise self.result_type(complete=True, passing=False, error=error, details=dict(runtime=runtime.dump()))
        if runtime.code != 0:
            error = Error(
                description=f"received status code {runtime.code}",
                suggestion="expected status code of zero")
            raise self.result_type(complete=True, passing=False, error=error, details=dict(runtime=runtime.dump()))
        return runtime



from typing import Any, Union, List
from decimal import Decimal

from ...grader.task import Result, Error, Message


class CorrectnessResult(Result):
    """The result of a correctness case."""

    expected: Any = None
    actual: Any = None

    def __init__(
            self,
            passing: bool,
            complete: bool = True,
            expected: Any = None,
            actual: Any = None,
            score: Union[Decimal, int, float, str] = None,
            error: Error = None,
            messages: List[Message] = None,
            details: dict = None):
        """Add fields for expected and actual."""

        super().__init__(
            passing=passing,
            complete=complete,
            score=score,
            error=error,
            messages=messages,
            details=details)
        self.expected = expected
        self.actual = actual

    def dump(self, thin: bool = False) -> dict:
        """Add extra details."""

        dump = super().dump()
        dump.update(expected=self.expected, actual=self.actual)
        return dump

    @classmethod
    def shorthand(
            cls,
            passing: bool = True,
            complete: bool = True,
            error: Error = None,
            actual: Any = None,
            expected: Any = None,
            **details) -> "CorrectnessResult":
        """Expand details."""

        return cls(
            complete=complete,
            passing=passing,
            error=error,
            actual=actual,
            expected=expected,
            details=details)




from curricula.log import log

from .task import Result
from .report import ProblemReport
from .task.dependency import fulfills_dependencies
from .task.registrar import TaskRegistrar
from .task.filter import filter_tasks
from .task.collection import TaskCollection
from ..resource import Context, Submission
from ..exception import GraderException

import typing

if typing.TYPE_CHECKING:
    from ..models import GradingProblem

__all__ = ("Grader",)


class Grader:
    """A main class for grading runtime."""

    register: TaskRegistrar

    # Assigned on import
    problem: "GradingProblem"

    def __init__(self):
        """Create the registrar."""

        self.register = TaskRegistrar()

    @property
    def tasks(self) -> TaskCollection:
        """Mirror register tasks."""

        return self.register.tasks

    def run(self, context: Context, submission: Submission) -> ProblemReport:
        """Build and test."""

        log.debug("setting up runtime")

        # Resources
        resources = dict(context=context, submission=submission, problem=self.problem)
        resources.update(resources=resources)

        # Create the filter
        filtered_task_names = filter_tasks(
            context=context,
            problem_short=self.problem.short,
            tasks=self.register.tasks)

        # Report
        report = ProblemReport.create(self.problem)

        # Execute
        for task in self.tasks:
            log.debug(f"running task {task.name}")

            # Check conditions for whether this case is filtered out, if so report is partial
            if task.name not in filtered_task_names:
                report.partial = True
                continue

            # If we can't run it, mark as incomplete
            elif not fulfills_dependencies(task, report):
                result = task.result_type.incomplete()

            # Run task if not hidden and dependencies are met
            else:
                try:
                    result = task.run(resources)

                # Results may be raised
                except Result as r:
                    result = r

                # Check if the result is the right type
                if type(result) is not task.result_type:
                    raise GraderException(
                        f"expected result type {task.result_type} from {task.name} in {task.source}")

            result.task = task
            report.automated.add(result)

        return report



import importlib.util
import json
import sys
from pathlib import Path
from typing import List
from dataclasses import field
from decimal import Decimal

from curricula.models import Assignment, Problem
from curricula.structure import Files
from .grader import Grader

__all__ = ("import_grader", "GradingProblem", "GradingAssignment")


def import_grader(grading_path: Path, problem: "GradingProblem", grader_name: str = "grader") -> Grader:
    """Import a grader from a tests file."""

    # Try to import as a module
    if grading_path.joinpath("__init__.py").is_file():
        name = f"_curricula_grading_{grading_path.parts[-1]}"
        spec = importlib.util.spec_from_file_location(name, str(grading_path.joinpath("__init__.py")))
        module = importlib.util.module_from_spec(spec)

    # Otherwise there's a tests.py
    else:
        name = f"_curricula_grading_{grading_path.parts[-1]}_tests"
        spec = importlib.util.spec_from_file_location(name, str(grading_path.joinpath("tests.py")))
        module = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)
    grader = getattr(module, grader_name)
    grader.problem = problem

    return grader


class GradingProblem(Problem):
    """Additional details for grading."""

    path: Path = field(init=False)
    grader: Grader = field(init=False)

    @property
    def automated_point_ratio(self) -> Decimal:
        return self.grading.automated.points / self.grader.weight()

    @classmethod
    def read(cls, data: dict, path: Path) -> "GradingProblem":
        """Import the grader."""

        self = GradingProblem.load(data)
        self.path = path
        if self.grading.is_automated:
            self.grader = import_grader(path, self)
        return self


class GradingAssignment(Assignment):
    """Additional details for grading."""

    path: Path

    problems: List[GradingProblem]

    @classmethod
    def read(cls, path: Path) -> "GradingAssignment":
        """Read the assignment.json and rebuild the models."""

        with path.joinpath(Files.INDEX).open("r") as file:
            data = json.load(file)

        problems = []
        for problem_data in data.pop("problems"):
            problems.append(GradingProblem.read(
                data=problem_data,
                path=path.joinpath(problem_data["short"])))

        assignment = GradingAssignment.load(data, problems=problems)
        assignment.path = path

        return assignment



import abc
from typing import Dict, TypeVar, Callable, Any, Type, Set, Optional, List, Union
from dataclasses import dataclass, field
from decimal import Decimal

from curricula.library.inject import inject
from curricula.log import log

__all__ = ("Message", "Result", "Dependencies", "Task", "Runnable", "Error")


@dataclass
class Message:
    """A message associated with a result."""

    class Kind:
        WARNING = "warning"
        INFO = "info"
        DEBUG = "debug"

    kind: str
    content: str

    def dump(self) -> dict:
        return dict(kind=self.kind, content=self.content)

    @classmethod
    def load(cls, data: dict) -> "Message":
        return cls(kind=data["kind"], content=data["content"])


@dataclass
class Error:
    """An error raised during a task."""

    description: str
    suggestion: str = None
    location: str = None
    traceback: str = None

    def dump(self, thin: bool = False) -> dict:
        if thin:
            return dict(description=self.description, suggestion=self.suggestion)
        return dict(
            description=self.description,
            suggestion=self.suggestion,
            location=self.location,
            traceback=self.traceback)

    @classmethod
    def load(cls, data: dict) -> "Error":
        return cls(**data)


@dataclass(init=False)
class Result(Exception, abc.ABC):
    """The result of a test."""

    complete: bool
    passing: bool
    details: dict
    error: Error
    messages: Optional[List[Message]]
    score: Decimal

    task: "Task" = field(init=False, repr=False)

    def __init__(
            self,
            passing: bool,
            complete: bool = True,
            score: Union[Decimal, int, float, str] = None,
            error: Error = None,
            messages: List[Message] = None,
            details: dict = None):
        """Initialize a new result.

        Details are passed as a dictionary in order to avoid potential
        collisions with normal arguments.
        """

        self.complete = complete
        self.passing = passing
        self.score = score
        self.error = error
        self.messages = messages if messages is not None else list()
        self.details = details if details is not None else dict()

    def dump(self, thin: bool = False) -> dict:
        """Serialize the result for JSON."""

        dump = dict(
            complete=self.complete,
            passing=self.passing,
            score=str(self.score) if self.score is not None else None,
            error=self.error.dump(thin=thin) if self.error is not None else None,
            messages=[message.dump() for message in self.messages],
            task=dict(name=self.task.name, description=self.task.description))
        if not thin:
            dump.update(details=self.details)
        return dump

    @classmethod
    def load(cls, data: dict, task: "Task") -> "Result":
        """Load a result from serialized."""

        data.pop("task")
        score = Decimal(score_data) if (score_data := data.pop("score", None)) is not None else None
        error = Error.load(error_data) if (error_data := data.pop("error", None)) is not None else None
        messages = list(map(Message.load, data.pop("messages")))
        self = cls(**data, score=score, error=error, messages=messages)
        self.task = task
        return self

    @classmethod
    def incomplete(cls):
        """Return a mock result if the task was not completed."""

        return cls(complete=False, passing=False)

    @classmethod
    def default(cls):
        """Called in special cases if no result is returned."""

        return cls(complete=True, passing=True)

    @classmethod
    def shorthand(cls, passing: bool = True, complete: bool = True, error: Error = None, **details):
        """Expand details."""

        cls(complete=complete, passing=passing, error=error, details=details)


TResult = TypeVar("TResult", bound=Result)
Runnable = Callable[[], TResult]


@dataclass
class Dependencies:
    """Task dependencies based on passing or completion."""

    passing: Set[str]
    complete: Set[str]

    def all(self) -> Set[str]:
        return self.passing.union(self.complete)

    @classmethod
    def normalize_from_details(cls, name: str, details: dict) -> Set[str]:
        """Normalize a set of strings."""

        value = details.pop(name, None)
        if value is None:
            return set()
        elif isinstance(value, str):
            return {value}
        elif isinstance(value, set):
            return value
        else:
            return set(value)

    @classmethod
    def from_details(cls, details: dict):
        """Parse from decorator details."""

        return cls(
            passing=cls.normalize_from_details("passing", details),
            complete=cls.normalize_from_details("complete", details))


@dataclass(eq=False)
class Task:
    """Superclass for check, build, run."""

    name: str
    description: str

    dependencies: Dependencies
    runnable: Runnable[Result]
    details: dict

    graded: bool
    weight: Optional[Decimal]
    points: Optional[Decimal]

    source: str
    tags: Set[str]

    result_type: Type[Result]

    def run(self, resources: Dict[str, Any]) -> Result:
        """Do the dependency injection for the runnable."""

        try:
            result = inject(resources, self.runnable)
        except ValueError as error:
            raise ValueError(f"caught in {self.name}: {error}")

        if result is None:
            log.debug(f"task {self.name} did not return a result")
            return self.result_type.default()

        return result



from typing import Optional

from curricula.library.valgrind import ValgrindReport

from ...grader.task import Result, Error


class MemoryResult(Result):
    """Result of a memory leak test."""

    kind = "memory"

    error_count: Optional[int]
    leaked_blocks: Optional[int]
    leaked_bytes: Optional[int]

    def __init__(
            self,
            passing: bool,
            complete: bool = True,
            error: Error = None,
            error_count: int = None,
            leaked_blocks: int = None,
            leaked_bytes: int = None,
            details: dict = None):
        """Initialize a memory result."""

        super().__init__(complete=complete, passing=passing, details=details)
        self.error_count = error_count
        self.leaked_blocks = leaked_blocks
        self.leaked_bytes = leaked_bytes

        if error is not None:
            self.error = error

        # Create the error with additional data
        elif error_count is not None and error_count > 0 or leaked_bytes is not None and leaked_bytes > 0:
            self.error = Error(description=f"leaked {leaked_bytes} bytes with {error_count} errors")

    @classmethod
    def from_valgrind_report(cls, report: ValgrindReport, **details) -> "MemoryResult":
        """Generate a memory test result from a valgrind report."""

        details.update(runtime=report.runtime.dump())

        if report.valgrind_errors is None:
            return cls(complete=False, passing=False, error=Error("failed to run valgrind"), details=details)

        leaked_blocks, leaked_bytes = report.memory_lost()
        return cls(
            passing=leaked_bytes == 0 and len(report.valgrind_errors) == 0,
            error_count=len(report.valgrind_errors),
            leaked_blocks=leaked_blocks,
            leaked_bytes=leaked_bytes,
            details=details)

    def dump(self, thin: bool = False) -> dict:
        dump = super().dump(thin=thin)
        dump.update(
            error_count=self.error_count,
            leaked_blocks=self.leaked_blocks,
            leaked_bytes=self.leaked_bytes)
        return dump



import abc
from typing import AnyStr, List, Sized, Union, Iterable, Collection, Callable, TypeVar, Sequence

from curricula.library.process import Interactive, TimeoutExpired, Interaction
from curricula.library.configurable import none, Configurable
from ...common import Evaluator, CompositeRunnable
from ...common.process import ProcessExecutor, ProcessStreamConnector, ProcessExitCodeConnector
from ...grader.task import Error
from . import CorrectnessResult

__all__ = (
    "as_lines",
    "lines_match",
    "lines_match_unordered",
    "CorrectnessRunnable",
    "CompareBytesEvaluator",
    "CompareExitCodeEvaluator",
    "ProcessCompareStreamTest",
    "ProcessCompareExitCodeTest",
    "write_then_read")


def as_lines(string: AnyStr) -> List[AnyStr]:
    """Strip and split by newline."""

    return string.strip().split("\n" if isinstance(string, str) else b"\n")


AnyStrSequence = Union[Iterable[AnyStr], Sized]


def lines_match(a: AnyStrSequence, b: AnyStrSequence) -> bool:
    """Check ordered equality.

    Returns a boolean indicating correctness and a list of errors
    encountered while checking.
    """

    if hasattr(a, "__len__") and hasattr(a, "__len__"):
        if len(a) != len(b):
            return False

    for x, y in zip(a, b):
        if x != y:
            return False

    return True


def lines_match_unordered(a: AnyStrSequence, b: AnyStrSequence) -> bool:
    """Check unordered equality."""

    return lines_match(sorted(a), sorted(b))


class CorrectnessRunnable(CompositeRunnable, metaclass=abc.ABCMeta):
    """Override annotation."""

    result_type = CorrectnessResult


BytesTransform = Callable[[bytes], bytes]
T = TypeVar("T")


def identity(x: T) -> T:
    return x


class CompareBytesEvaluator(Configurable, Evaluator):
    """Compares output to expected values."""

    out_transform: BytesTransform
    out_line_transform: BytesTransform
    test_out: bytes
    test_out_lines: Iterable[bytes]
    test_out_lines_lists: Iterable[Iterable[bytes]]

    def __init__(
            self,
            *,
            out_transform: BytesTransform = none,
            out_line_transform: BytesTransform = none,
            test_out: bytes = none,
            test_out_lines: Iterable[bytes] = none,
            test_out_lines_lists: Iterable[Iterable[bytes]] = none,
            **kwargs):
        """Build a test for matching stdout output.

        The list of lines from the stdout of the runtime is compared
        against each list of lines in test_out_line_lists. Note that this
        method first checks whether any error was raised during runtime.
        """

        super().__init__(**kwargs)

        self.test_out = self.resolve("test_out", local=test_out, default=None)
        self.test_out_lines = self.resolve("test_out_lines", local=test_out_lines, default=None)
        self.test_out_lines_lists = self.resolve("test_out_lines_lists", local=test_out_lines_lists, default=None)

        # Check resolvable
        resolvable = tuple(filter(None, (self.test_out, self.test_out_lines, self.test_out_lines_lists)))
        if len(resolvable) != 1:
            raise ValueError("Exactly one of test_out, test_out_lines, or test_out_lines_lists is required")

        self.out_transform = self.resolve("out_transform", local=out_transform, default=identity)
        self.out_line_transform = self.resolve("out_line_transform", local=out_line_transform, default=identity)

    def evaluate(self, out: bytes) -> CorrectnessResult:
        """Call the corresponding test."""

        if self.test_out is not None:
            return self._test_out(out)
        return self._test_out_lines_lists(out)

    def _test_out(self, out: bytes) -> CorrectnessResult:
        """Shortcut compare for one option block text."""

        test_out = self.resolve("test_out")

        out = self.out_transform(out)
        if out == test_out:
            return CorrectnessResult.shorthand()

        return CorrectnessResult.shorthand(
            passing=False,
            error=Error(description="unexpected output"),
            expected=test_out.decode(errors="replace"),
            actual=out.decode(errors="replace"))

    def _test_out_lines_lists(self, out: bytes) -> CorrectnessResult:
        """Used to compare multiple options of lists of lines."""

        if self.test_out_lines is not None:
            test_out_lines_lists = [self.test_out_lines]
        else:
            test_out_lines_lists = self.test_out_lines_lists

        out_lines = tuple(map(self.out_line_transform, self.out_transform(out).split(b"\n")))

        if any(lines_match(out_lines, lines) for lines in test_out_lines_lists):
            return CorrectnessResult.shorthand()

        expected = tuple(b"\n".join(test_out_lines).decode(errors="replace") for test_out_lines in test_out_lines_lists)
        return CorrectnessResult(
            passing=False,
            error=Error(description="unexpected output"),
            actual=b"\n".join(out_lines).decode(errors="replace"),
            expected=expected[0] if len(expected) == 1 else expected)


class CompareExitCodeEvaluator(Evaluator, Configurable):
    """Checks program exit code."""

    expected_code: int
    expected_codes: Collection[int]

    _expected_codes: Collection[int]

    def __init__(self, *, expected_code: int = none, expected_codes: Collection[int] = none, **kwargs):
        """Set expected codes."""

        super().__init__(**kwargs)

        expected_code = self.resolve("expected_code", local=expected_code, default=None)
        expected_codes = self.resolve("expected_codes", local=expected_codes, default=None)
        if expected_code is None and expected_codes is None:
            raise ValueError("Runtime exit test requires either expected status or statuses!")
        if expected_code is not None and expected_codes is not None:
            raise ValueError("Runtime exit test requires either expected status or statuses!")

        self._expected_codes = (expected_code,) if expected_code is not None else expected_codes

    def evaluate(self, code: int) -> CorrectnessResult:
        """Check if the exit code is one of the options."""

        if code in self._expected_codes:
            return CorrectnessResult(passing=True)
        return CorrectnessResult(
            passing=False,
            error=Error(
                description=f"received incorrect exit code {code}",
                suggestion=f"""expected exit code {", ".join(map(str, self._expected_codes))}"""))


class ProcessCompareStreamTest(
        ProcessExecutor,
        ProcessStreamConnector,
        CompareBytesEvaluator,
        CorrectnessRunnable):
    """Standard process output test."""


class ProcessCompareExitCodeTest(
        ProcessExecutor,
        ProcessExitCodeConnector,
        CompareExitCodeEvaluator,
        CorrectnessRunnable):
    """Make sure the process exits as expected."""


def write_then_read(
        interactive: Interactive,
        stdin: Sequence[bytes],
        read_condition: Callable[[bytes], bool] = None,
        read_condition_timeout: float = None) -> Interaction:
    """Pass in a bunch of lines, get the last output, compare."""

    if not interactive.poll():
        raise CorrectnessResult(passing=False, error=Error(description="program crashed"))

    with interactive.recording() as interaction:
        for command in stdin:
            interactive.stdin.write(command)
            try:
                interactive.stdout.read(
                    condition=read_condition,
                    timeout=read_condition_timeout)
            except TimeoutExpired:
                raise CorrectnessResult.shorthand(
                    passing=False,
                    error=Error(description="timed out"),
                    runtime=interaction.dump())

    if interaction.stdout is None:
        raise CorrectnessResult.shorthand(
            passing=False,
            error=Error(description="did not receive output"),
            runtime=interaction.dump())

    return interaction



from curricula.library.valgrind import ValgrindReport, run
from ..correctness.common import ProcessExecutor
from . import MemoryResult


class MemoryTest(ProcessExecutor):
    """Runs valgrind on the executable with given parameters."""

    def __call__(self, resources: dict) -> MemoryResult:
        """Check if the output matches."""

        self.resources = resources
        self.details = {}
        result = MemoryResult.from_valgrind_report(self.execute())
        result.details.update(self.details)
        self.resources = None
        self.details = None
        return result

    def execute(self) -> ValgrindReport:
        executable = self.resources[self.executable_name]
        report = run(
            *executable.args,
            *self.resolve("args", default=()),
            stdin=self.resolve("stdin", default=None),
            timeout=self.resolve("timeout", default=None),
            cwd=self.resolve("cwd", default=None))
        return report


class GoogleMemoryTest(MemoryTest):
    """Custom gtest args."""

    def __init__(self, test_name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = (test_name,)



import json
import statistics
from typing import List, Dict, Iterable, Union, Set
from dataclasses import dataclass, field
from pathlib import Path

from ..models import GradingAssignment
from ..grader.report import AssignmentReport
from ..grader.task import Task


@dataclass
class TaskSummary:
    """Statistics about task results."""

    task: Task
    students_complete: List[dict] = field(default_factory=list)
    students_passing: List[dict] = field(default_factory=list)
    students_timeout: List[dict] = field(default_factory=list)


@dataclass
class StudentProblemSummary:
    """Problem results for a single student."""

    tasks_complete: List[Task] = field(default_factory=list)
    tasks_passing: List[Task] = field(default_factory=list)


@dataclass
class StudentSummary:
    """Statistics for an individual student."""

    student: dict
    problems: Dict[str, StudentProblemSummary]

    def __init__(self, student: dict, problem_shorts: Iterable[str]):
        self.student = student
        self.problems = {}
        for problem_short in problem_shorts:
            self.problems[problem_short] = StudentProblemSummary()


@dataclass
class ProblemSummary:
    """Statistics about a set of tests."""

    tasks: Dict[str, TaskSummary]
    failed_setup: Set[str]

    def __init__(self, tasks: Iterable[Task]):
        self.tasks = {task.name: TaskSummary(task) for task in tasks}
        self.failed_setup = set()


@dataclass
class Summary:
    """Overall assignment statistics."""

    problems: Dict[str, ProblemSummary]
    students: Dict[str, StudentSummary]
    failed_setup: Set[str]

    def __init__(self, assignment: GradingAssignment, students: dict):
        """Generate skeleton of data store."""

        self.problems = {}
        for problem in assignment.problems:
            if problem.grading.is_automated:
                self.problems[problem.short] = ProblemSummary(problem.grader.tasks)
        self.students = {}
        problem_shorts = tuple(self.problems.keys())
        for student_username, student in students.items():
            self.students[student_username] = StudentSummary(student, problem_shorts)
        self.failed_setup = set()


def build_student_summary(summary: Summary, assignment: GradingAssignment, report_path: Path):
    """Build a table of results with axes students and tasks."""

    report_name = report_path.parts[-1].rsplit(".")[0]
    student_summary = summary.students[report_name]

    with open(report_path) as file:
        report = AssignmentReport.load(json.load(file), assignment)

    for problem_short, problem_report in report.problems.items():
        problem_summary = summary.problems[problem_short]

        for task_name, result in problem_report.results.items():
            task_summary = problem_summary.tasks[task_name]
            task = task_summary.task
            if task.stage == "setup" and not result.passing:
                summary.failed_setup.add(report_name)
                problem_summary.failed_setup.add(report_name)

            if result.complete:
                task_summary.students_complete.append(student_summary.student)
                student_summary.problems[problem_short].tasks_complete.append(task)
            if result.passing:
                task_summary.students_passing.append(student_summary.student)
                student_summary.problems[problem_short].tasks_passing.append(task)
            if "runtime" in result.details and result.details["runtime"] is not None:
                if result.details["runtime"]["timed_out"]:
                    task_summary.students_timeout.append(student_summary.student)

    return student_summary


def build_summary(assignment: GradingAssignment, report_paths: Iterable[Path]) -> Summary:
    """Compile problem summaries."""

    students = {}
    for report_path in report_paths:
        student_username = report_path.parts[-1].rsplit(".")[0]
        students[student_username] = dict(username=student_username)
    summary = Summary(assignment, students)
    for report_path in report_paths:
        build_student_summary(summary, assignment, report_path)
    return summary


def percent(x: Union[int, float], n: int = 1) -> str:
    return f"{round(x / n * 1000) / 10 if n != 0 else 0}%"


def filter_tests(tasks: List[Task]) -> List[Task]:
    return [task for task in tasks if task.stage == "test"]


def summarize(grading_path: Path, report_paths: Iterable[Path]):
    """Do summaries of the reports in a directory."""

    assignment = GradingAssignment.read(grading_path)
    summary = build_summary(assignment, report_paths)

    for problem_short, problem_summary in summary.problems.items():
        print(f"Problem: {problem_short}")

        print("  Tests")
        for task_name, task_summary in problem_summary.tasks.items():
            print(f"    {task_name}: {len(task_summary.students_passing)}/{len(task_summary.students_complete)}",
                  f"({len(task_summary.students_timeout)} timeout)")

        scores = []
        for student_username, student_summary in summary.students.items():
            count_tests_complete = len(filter_tests(student_summary.problems[problem_short].tasks_complete))
            if count_tests_complete:
                count_tests_passing = len(filter_tests(student_summary.problems[problem_short].tasks_passing))
                scores.append(count_tests_passing / count_tests_complete)
                # print(student_username, 100 * count_tests_passing / count_tests_complete)

        print("  Statistics")
        print(f"    Total scores: {len(scores)}")
        print(f"    Mean: {percent(statistics.mean(scores)) if len(scores) > 0 else '-'}")
        print(f"    Median: {percent(statistics.median(scores)) if len(scores) > 0 else '-'}")
        print(f"    Perfect: {percent(len(list(filter(lambda x: x == 1, scores))), len(scores))}")
        print(f"    Zeros: {percent(len(list(filter(lambda x: x == 0, scores))), len(scores))}")
        print(f"    Failed setup: {len(problem_summary.failed_setup)}")

    print(f"Total submissions that failed setup: {len(summary.failed_setup)}")
    for report_name in summary.failed_setup:
        print(f"  {report_name}")



import functools
from typing import Type, Any, Set, Optional, Callable
from decimal import Decimal

from curricula.library.debug import get_source_location

from . import Runnable, Task, Result, Dependencies
from .collection import TaskCollection
from ...exception import GraderException


def is_some(value: Any) -> bool:
    """Shorthand for checking if identical to None."""

    return value is not None


def first_some(*args: Any) -> Any:
    """Return the first arg that is not None."""

    try:
        return next(filter(None, args))
    except StopIteration:
        return None


def underwrite(source: dict, destination: dict) -> dict:
    """Fill keys in source not defined in destination."""

    for key, value in source.items():
        if key not in destination:
            destination[key] = value
    return destination


class TaskRegistrar:
    """Utility class for registering tasks."""

    tasks: TaskCollection

    def __init__(self):
        """Assign task collection and dynamically create shortcuts."""

        self.tasks = TaskCollection()

    def __call__(self, runnable: Any = None, /, **details) -> Optional[Callable[[Runnable], None]]:
        """Explicitly register a runnable."""

        if runnable is not None:
            self.register(runnable=runnable, details=details, result_type=self._determine_result_type(runnable))
            return

        def register(r: Runnable, /):
            self.register(runnable=r, details=details, result_type=self._determine_result_type(r))
        return register

    def __getitem__(self, result_type: Type[Result]):
        """Partial in the result_type."""

        def register(runnable: Runnable = None, /, **details):
            """Return potentially a decorator or maybe not."""

            if runnable is not None:
                self.register(runnable=runnable, details=details, result_type=result_type)
                return

            return functools.partial(self.register, details=details, result_type=result_type)

        return register

    def register(self, runnable: Any, details: dict, result_type: Type[Result]):
        self.tasks.push(Task(
            name=TaskRegistrar._resolve_name(runnable, details),
            description=TaskRegistrar._resolve_description(runnable, details),
            dependencies=Dependencies.from_details(details),
            runnable=runnable() if isinstance(runnable, type) else runnable,
            graded=details.pop("graded", True),
            weight=Decimal(w) if is_some(w := details.pop("weight", None)) else None,
            points=Decimal(p) if is_some(p := details.pop("points", None)) else None,
            source=get_source_location(1),
            tags=TaskRegistrar._resolve_tags(details),
            result_type=result_type,
            details=details))

    def weight(self) -> Decimal:
        """Combined weight of all graded tasks."""

        return sum(map(lambda t: t.weight, filter(lambda t: t.graded, self.tasks)), Decimal(0))

    @staticmethod
    def _determine_result_type(runnable: Runnable) -> Type[Result]:
        result_type_candidate = None
        if is_some(annotations := getattr(runnable, "__annotations__", None)):
            result_type_candidate = annotations.get("return")
        if result_type_candidate is None:
            result_type_candidate = getattr(runnable, "result_type", None)
        if result_type_candidate is None:
            raise GraderException("result type must be specified by return annotation or bracket notation")
        if not issubclass(result_type_candidate, Result):
            raise GraderException("annotated result type does not inherit from Result")
        return result_type_candidate

    @staticmethod
    def _resolve_name(runnable: Runnable, details: dict) -> str:
        if is_some(name := details.pop("name", None)):
            return name
        if is_some(name := getattr(runnable, "__qualname__", None)):
            return name
        raise GraderException("No viable candidate for task name, please provide during registration")

    @staticmethod
    def _resolve_description(runnable: Runnable, details: dict) -> Optional[str]:
        return details.pop("description", None) or runnable.__doc__

    @staticmethod
    def _resolve_tags(details: dict) -> Set[str]:
        if isinstance(tags := details.pop("tags", set()), set):
            return tags
        raise GraderException("tags must either be a set of strings or None")



from ..grader.task import Result, Error


class SetupResult(Result):
    """Common result for generic tasks."""

    def __init__(self, passing: bool = True, complete: bool = True, error: Error = None, **details):
        super().__init__(complete=complete, passing=passing, error=error, details=details)



import timeit

from pathlib import Path
from typing import Iterator, Iterable, Tuple

from curricula.log import log
from .models import GradingAssignment
from .grader.report import AssignmentReport
from .resource import Submission, Context


def _run(assignment: GradingAssignment, submission_path: Path, options: dict) -> AssignmentReport:
    """Actual runtime."""

    report = AssignmentReport.create(assignment)
    context = Context(options=options)
    for problem in filter(lambda p: p.grading.is_automated, assignment.problems):
        log.debug(f"running problem {problem.short}")
        submission = Submission(
            assignment_path=submission_path,
            problem_path=submission_path.joinpath(problem.relative_path))
        report[problem.short] = problem.grader.run(context=context, submission=submission)
    return report


def run(assignment: GradingAssignment, submission_path: Path, options: dict) -> AssignmentReport:
    """Run all tests on a submission and return a dict of results."""

    log.info(f"running {submission_path}")
    start = timeit.default_timer()
    report = _run(assignment, submission_path, options)
    elapsed = timeit.default_timer() - start
    log.debug(f"finished in {round(elapsed, 5)} seconds")
    return report


def run_batch(
        assignment: GradingAssignment,
        submission_paths: Iterable[Path],
        options: dict) -> Iterator[Tuple[Path, AssignmentReport]]:
    """Run multiple reports, map directory to report."""

    # Start timer
    start = timeit.default_timer()

    for submission_path in submission_paths:
        yield submission_path, run(assignment, submission_path, options=options)

    elapsed = timeit.default_timer() - start
    log.info(f"finished batch in {round(elapsed, 5)} seconds")
